// Copyright 2002-2003 University of Guelph, All Rights Reserved
import java.util.*;
import java.io.*;

class MutateXover {

  public static final int POPULATION_SIZE = 400;
  public static final int NUM_UNFIT = 125;

  public TreeSet population;
  public Random mutateRandom;
  public Problem problem;
  public static Vector rows;
  public ArrayList parents;
  public ArrayList unfitIndividuals;

  public MutateXover(Problem problemObject) {
	problem = problemObject;
	initPopulation();
  }

  public void findSolution() {   
    Xover.numCases = 0;
    evolveSolution();
  }

  public void evolveSolution() {	
	mutateRandom = new Random();
	fitnessSort();

	while (bestNotSolution()) {
	  select();
	  reproduce();
	  fitnessSort();

	  if (Debug.DEBUG_LEVEL >= Debug.DEBUG_QUIET) {
		System.out.println("Tested " + String.valueOf(Xover.numCases) + 
						 " possible solutions; Top fitness: " +
						 String.valueOf(((XoverChromosone)(population.first())).getFitness()));	  
	  }
	 
	  if (Debug.DEBUG_LEVEL >= Debug.DEBUG_MEDIUM) {
		Object popArray[] = population.toArray();
		for (int i = 0; i < 6; i++) {
		  System.out.println((XoverChromosone)(popArray[i]));
		  if (Debug.DEBUG_LEVEL >= Debug.DEBUG_HIGH) {
			System.out.println(((XoverChromosone)(popArray[i])).getSolutionString());
		  }
		}
		//		for (int i = popArray.length - 5; i < popArray.length; i++) {
		//		  System.out.println(((XoverChromosone)(popArray[i])));
		//		}		
	  }
	  problem.solution = ((XoverChromosone)(population.first())).getSolution();
	}
  }

  public boolean bestNotSolution() {
	boolean isSolution = (((XoverChromosone)(population.first())).isSolution());

	if (!isSolution) {
	  if (Debug.DEBUG_LEVEL >= Debug.DEBUG_HIGH) {
		System.out.print(" " + String.valueOf(problem.checkConflicts()) + " ");
	  }
	}
	return !isSolution;
  }

  public void fitnessSort() {
	//	Arrays.sort(population);
  }

  public int getParent(double chanceRepro[]) {
	double randParent = mutateRandom.nextDouble();

	int i = 0;
	double maxChance = chanceRepro[i];
	
	while(randParent > maxChance) {
	  i++;
	  maxChance += chanceRepro[i];
	}
	return i;
  }

  public void select() {
	Object popArray[] = population.toArray();
	double bottomFitness = ((XoverChromosone)(popArray[POPULATION_SIZE - NUM_UNFIT - 1])).getFitness();
	double chanceRepro[] = new double[POPULATION_SIZE - NUM_UNFIT];
	double timesBetter[] = new double[POPULATION_SIZE - NUM_UNFIT];
	double totalTimesBetter = 0;
	int parentNum;

	for (int i = 0; i < POPULATION_SIZE - NUM_UNFIT; i++) {
	  timesBetter[i] = (10000 - ((XoverChromosone)(popArray[i])).getFitness()) / (10000 - bottomFitness);
	  totalTimesBetter += timesBetter[i];
	}
	double x = 1 / totalTimesBetter;

	for (int i = 0; i < POPULATION_SIZE - NUM_UNFIT; i++) {
	  chanceRepro[i] = timesBetter[i] * x;
	}

	parents = new ArrayList();
	unfitIndividuals = new ArrayList();
	int parent2Num;

	int i;
	for (i = 0; i < (int)(NUM_UNFIT / 2); i++) {
	  parentNum = getParent(chanceRepro);
	  parents.add(popArray[parentNum]);
	  parent2Num = (POPULATION_SIZE - NUM_UNFIT) - parentNum;
	  parents.add(popArray[parent2Num]);
	}
	if (i < NUM_UNFIT + 1) {
	  parents.add(popArray[getParent(chanceRepro)]);
	}

	for (i = population.size() - NUM_UNFIT ; i < population.size(); i++) {
	  unfitIndividuals.add(popArray[i]);
	}

  }

  public void reproduce() {

	XoverChromosone individual;
	Iterator parentIterator = parents.iterator();
	XoverChromosone parent1;
	XoverChromosone parent2;
	XoverChromosone sibling1;
	XoverChromosone sibling2;

	parent2 = null;

	while (parentIterator.hasNext()) {
	  sibling1 = new XoverChromosone(problem);
	  sibling2 = new XoverChromosone(problem);
	  parent1 = (XoverChromosone)(parentIterator.next());
	  if (parentIterator.hasNext()) {
		parent2 = (XoverChromosone)(parentIterator.next());
	  }
	  sibling1.reproduce(parent1, parent2, sibling2, mutateRandom);	 
	  Xover.numCases++;
	  do {
		sibling1.mutate(mutateRandom, problem.getMaxVars());
	  } while (population.contains(sibling1));
	  population.add(sibling1);	  
	  do {
		sibling2.mutate(mutateRandom, problem.getMaxVars());
	  } while (population.contains(sibling2));
	  population.add(sibling2);	  
	}

	for (int i = 0; i < unfitIndividuals.size(); i++) {
	  population.remove(unfitIndividuals.get(i));	  
	}
	unfitIndividuals = null;

  }

  public void initPopulation() {
	int maxDomain = problem.getMaxDomain();
	Vector varPermutations;
	TreeSet topDomain = new TreeSet();
	Vector varSolution;
	Random permRandom;

	for (int i = 1; i <= maxDomain; i++) {
	  topDomain.add(new Integer(i));
	}

	varPermutations = doPermute(topDomain);
	rows = varPermutations;

	population = new TreeSet();

	permRandom = new Random();
	int permChoice;
	int solution[];
	XoverChromosone individual;

	solution = new int[problem.getMaxVars() + 1];
	
	for (int i = 0; i < POPULATION_SIZE; i++) {
	  Xover.numCases++;
	  for (int j = 0; j < problem.getMaxDomain(); j++) {
		permChoice = permRandom.nextInt(varPermutations.size());
		setSolution(varPermutations.elementAt(permChoice), j * problem.getMaxDomain() + 1, solution);
	  }		
	  individual = new XoverChromosone(solution, problem);
	  while (population.contains(individual)) {
		for (int j = 0; j < problem.getMaxDomain(); j++) {
		  permChoice = permRandom.nextInt(varPermutations.size());
		  setSolution(varPermutations.elementAt(permChoice), j * problem.getMaxDomain() + 1, solution);
		}
		individual = new XoverChromosone(solution, problem);
	  }	 
	  population.add(individual);			
	}
  }

  public static void setSolution(Object permVector, int varStart, int solution[]) {
	Vector permutes = (Vector)(permVector);
	for (int o = 0; o < permutes.size(); o++) {
	  solution[varStart + o] = 
		((Integer)(permutes.elementAt(o))).intValue();
	}
  }

  public static Vector doPermute(TreeSet curDomain) {
	Vector permutations = new Vector();
	Vector integers;

	if (curDomain.size() == 1) {
	  integers = new Vector();
	  integers.addElement((Integer)(curDomain.first()));
	  permutations.addElement(integers);
	} else {	  
	  Iterator current = curDomain.iterator();	 
	  while (current.hasNext()) {
		TreeSet newDomain = (TreeSet)(curDomain.clone());
		Integer curInt = (Integer)(current.next());
		newDomain.remove(curInt);

		Vector endCombos = doPermute(newDomain);
		for (int i = 0; i < endCombos.size(); i++) {
		  integers = new Vector();
		  integers.addElement(curInt);
		  Vector endInts = (Vector)(endCombos.elementAt(i));

		  for (int j = 0; j < endInts.size(); j++) {
			integers.addElement((Integer)(endInts.elementAt(j)));
		  }
		  permutations.addElement(integers);
		}		
	  }	  
	}

	if (Debug.DEBUG_LEVEL >= Debug.DEBUG_VERBOSE) {
	  for (int i = 0; i < permutations.size(); i++) {
		Vector ints = (Vector)(permutations.elementAt(i));
		System.out.print(String.valueOf(i) + ": ");
		for (int j = 0; j < ints.size(); j++) {
		  System.out.print(String.valueOf((Integer)(ints.elementAt(j)) + " "));
		}
		System.out.println("");
	  }
	}
	return permutations;
  }
}

